import ipaddress
import overseer
from threading import Thread
from io import open
import socket
import os
import re
import time


class ScanManager:
    def __init__(self):
        self.pending_shutdown = False
        self.active_scans = []
        self.broadcast_thread = Thread(target=self.__broadcast_thread)
        self.broadcast_thread.start()

    def __broadcast_thread(self):
        while not self.pending_shutdown:
            time.sleep(5)
            if len(self.active_scans) == 0:
                continue

            for scan in self.active_scans:
                # WebSocket progress
                total_progress = (scan.tcp_progress + scan.udp_progress) / 2
                overseer.api.send_websocket_event(
                    {
                        "id": scan.history_id,
                        "target": scan.target,
                        "status": "RUNNING",
                        "tcp_progress": scan.tcp_progress,
                        "udp_progress": scan.udp_progress,
                        "total_progress": round(total_progress),
                    }
                )

                if scan.tcp_progress + scan.udp_progress != 200:
                    continue

                # Combine ports
                results = list(map(lambda x: "%s UDP" % x, scan.udp_results))
                results.extend(
                    list(map(lambda x: "%s TCP" % x, scan.tcp_results))
                )  # noqa: E501
                results.sort()

                # Update database
                overseer.database.update_scan_result(
                    scan.history_id, "COMPLETE", results=results
                )

                # WebSocket completion
                overseer.api.send_websocket_event(
                    {
                        "id": scan.history_id,
                        "target": scan.target,
                        "status": "COMPLETE",
                        "results": results,
                    }
                )

                # Cleanup active scan
                scan.join()
                self.active_scans.remove(scan)

    def shutdown(self):
        self.pending_shutdown = True
        self.broadcast_thread.join()

    def get_status(self):
        return list(
            map(
                lambda x: {
                    "id": x.history_id,
                    "target": x.target,
                    "status": "IN_PROGRESS",
                    "tcp_progress": x.tcp_progress,
                    "udp_progress": x.udp_progress,
                    "total_progress": round(
                        (x.tcp_progress + x.udp_progress) / 2
                    ),  # noqa: E501
                },
                self.active_scans,
            )
        )

    def perform_scan(self, target):
        try:
            target = socket.gethostbyname(target)
        except socket.error:
            return None

        if self.is_ip(target):
            scan_history = overseer.database.create_scan_result(
                "IN_PROGRESS", ip_addr=target
            )
        else:
            scan_history = overseer.database.create_scan_result(
                "IN_PROGRESS", hostname=target
            )

        new_scan = Scanner(target, scan_history.id)
        new_scan.start()
        self.active_scans.append(new_scan)
        return scan_history.id

    def is_ip(self, target):
        try:
            ipaddress.ip_address(target)
            return True
        except ValueError:
            return False


class Scanner(Thread):
    def __init__(self, target, history_id):
        Thread.__init__(self)
        self.target = target
        self.history_id = history_id

        self.port_count = 1000

        self.tcp_progress = 0
        self.udp_progress = 0

        self.tcp_results = []
        self.udp_results = []

        self.udp_payloads = {}
        self.__load_nmap_payloads()

    def __load_nmap_payloads(self):
        """Load and parse nmap UDP payloads"""

        # Open file & remove comments
        nmap_payloads = os.path.join(
            os.path.dirname(__file__), "./resources/nmap-payloads"
        )
        f = open(nmap_payloads, "r", encoding="unicode_escape")
        raw_file = re.sub(
            r"^(#|\s*#).*$", "", f.read(), flags=re.MULTILINE | re.UNICODE
        )
        f.close()

        # Find all matches
        results = re.findall(
            r"(^udp.*?)(?=udp|\Z)",
            raw_file,
            flags=re.MULTILINE | re.DOTALL | re.UNICODE,
        )

        for raw_match in results:
            match = raw_match.strip()
            match_payloads = list(
                map(lambda x: x.strip(), re.findall(r'\"(.*)"', match))
            )
            raw_ports = re.match(r"^udp\s([\d,-]*)", match).group(1)

            for raw_port in raw_ports.split(","):
                if "-" in raw_port:
                    port_range = raw_port.split("-")
                    start_port = int(port_range[0])
                    end_port = int(port_range[1])
                    for port_match in range(start_port, end_port + 1):
                        if port_match not in self.udp_payloads:
                            self.udp_payloads[port_match] = []
                        self.udp_payloads[port_match].extend(match_payloads)
                else:
                    port_match = int(raw_port)
                    if port_match not in self.udp_payloads:
                        self.udp_payloads[port_match] = []
                    self.udp_payloads[port_match].extend(match_payloads)

    def run(self):
        tcp_thread = Thread(target=self.__scan_tcp)
        udp_thread = Thread(target=self.__scan_udp)
        tcp_thread.start()
        udp_thread.start()
        tcp_thread.join()
        udp_thread.join()
        return {"TCP": self.tcp_results, "UDP": self.udp_results}

    def __scan_tcp(self):
        for port in range(1, self.port_count):
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.settimeout(0.1)
            result = s.connect_ex((self.target, port))
            if result == 0:
                self.tcp_results.append(port)
            s.close()
            self.tcp_progress = round(port / self.port_count * 100)

    def __scan_udp(self):
        for port in range(1, self.port_count):
            # print("UDP port %s..." % port)
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            s.settimeout(0.01)
            payloads = ["\x00"]
            if port in self.udp_payloads:
                payloads = self.udp_payloads[port]
            for payload in payloads:
                s.sendto(payload.encode("utf-8"), (self.target, port))
                try:
                    s.recvfrom(1)
                    self.udp_results.append(port)
                    break
                except ConnectionRefusedError:
                    pass
                except socket.timeout:
                    pass
            s.close()
            self.udp_progress = round(port / self.port_count * 100)


# FOR TESTING PURPOSES
def main():
    sm = ScanManager()
    sm.perform_scan("localhost")
    # sm.perform_scan("10.0.20.254")
    # sm.perform_scan("10.0.21.20")


if __name__ == "__main__":
    main()
